#include "Visu.h"

#include <algorithm>
#include <exception>
#include <string>

#include <Eigen/Geometry>

#include <SimoxUtility/algorithm/get_map_keys_values.h>
#include <SimoxUtility/math/pose.h>
#include <SimoxUtility/math/rescale.h>

#include <VirtualRobot/XML/RobotIO.h>

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/time/CycleUtil.h>
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/interface/core/PackagePath.h>

#include <RobotAPI/libraries/armem/core/Time.h>

#include <RobotAPI/components/ArViz/Client/Elements.h>

#include <RobotAPI/libraries/armem_robot_state/server/description/Segment.h>
#include <RobotAPI/libraries/armem_robot_state/server/localization/Segment.h>
#include <RobotAPI/libraries/armem_robot_state/server/proprioception/Segment.h>
#include <RobotAPI/libraries/armem_robot_state/server/proprioception/SensorValues.h>

#include "combine.h"


namespace armarx::armem::server::robot_state
{

    Visu::Visu(
        const description::Segment& descriptionSegment,
        const proprioception::Segment& proprioceptionSegment,
        const localization::Segment& localizationSegment)
        : descriptionSegment(descriptionSegment),
          proprioceptionSegment(proprioceptionSegment),
          localizationSegment(localizationSegment)
    {
        Logging::setTag("Visu");
    }


    void Visu::defineProperties(armarx::PropertyDefinitionsPtr defs, const std::string& prefix)
    {
        defs->optional(p.enabled, prefix + "enabled", "Enable or disable visualization of objects.");
        defs->optional(p.frequencyHz, prefix + "frequenzyHz", "Frequency of visualization.");
        defs->optional(p.framesEnabled, prefix + "framesEnabled", "Enable or disable visualization of frames.");
        defs->optional(p.forceTorque.enabled, prefix + "forceTorque.enabled",
                       "Enable or disable visualization of force torque sensors.");
        defs->optional(p.forceTorque.forceScale, prefix + "forceTorque.forceScale",
                       "Scaling of force arrows.");
    }


    void Visu::init()
    {
    }


    void Visu::connect(const viz::Client& arviz, DebugObserverInterfacePrx debugObserver)
    {
        this->arviz = arviz;
        if (debugObserver)
        {
            bool batchMode = true;
            this->debugObserver = DebugObserverHelper("RobotStateMemory", debugObserver, batchMode);
        }

        if (updateTask)
        {
            updateTask->stop();
            updateTask->join();
            updateTask = nullptr;
        }
        updateTask = new SimpleRunningTask<>([this]()
        {
            this->visualizeRun();
        });
        updateTask->start();
    }


    void Visu::visualizeRobots(viz::Layer& layer, const robot::Robots& robots)
    {
        for (const robot::Robot& robot : robots)
        {
            const armarx::data::PackagePath xmlPath = robot.description.xml.serialize();

            // clang-format off
            viz::Robot robotVisu = viz::Robot(robot.description.name)
                                   .file(xmlPath.package, xmlPath.package + "/" + xmlPath.path)
                                   .joints(robot.config.jointMap)
                                   .pose(robot.config.globalPose);

            robotVisu.useFullModel();
            // clang-format on

            layer.add(robotVisu);
        }
    }


    void Visu::visualizeFrames(
        viz::Layer& layerFrames,
        const std::unordered_map<std::string, std::vector<Eigen::Affine3f>>& frames)
    {
        for (const auto& [robotName, robotFrames] : frames)
        {
            Eigen::Affine3f pose = Eigen::Affine3f::Identity();

            int i = 0;
            for (const auto& frame : robotFrames)
            {
                Eigen::Affine3f from = pose;
                Eigen::Affine3f to = pose * frame;

                pose = to;

                layerFrames.add(viz::Arrow(robotName + std::to_string(++i)).fromTo(from.translation(), to.translation()));
            }
        }
    }

    void Visu::visualizeFramesIndividual(
        viz::Layer& layerFrames,
        const std::unordered_map<std::string, std::vector<Eigen::Affine3f>>& frames)
    {

        std::vector<std::string> FRAME_NAMES{/*world*/ "map", "odom", "robot"};

        for (const auto& [robotName, robotFrames] : frames)
        {
            int i = 0;
            for (const auto& frame : robotFrames)
            {
                layerFrames.add(viz::Pose(robotName + FRAME_NAMES.at(i++)).pose(frame.matrix()).scale(3));
            }
        }
    }



    void Visu::visualizeRun()
    {
        CycleUtil cycle(static_cast<int>(1000 / p.frequencyHz));
        while (updateTask and not updateTask->isStopped())
        {
            if (p.enabled)
            {
                const Time timestamp = Time::Now();
                ARMARX_DEBUG << "Visu task at " << armem::toStringMilliSeconds(timestamp);

                try
                {
                    visualizeOnce(timestamp);
                }
                catch (const std::exception& e)
                {
                    ARMARX_WARNING << "Caught exception while visualizing robots: \n" << e.what();
                }
                catch (...)
                {
                    ARMARX_WARNING << "Caught unknown exception while visualizing robots.";
                }

                if (debugObserver.has_value())
                {
                    debugObserver->sendDebugObserverBatch();
                }
            }
            cycle.waitForCycleDuration();
        }
    }


    void Visu::visualizeOnce(const Time& timestamp)
    {
        TIMING_START(tVisuTotal);

        // TODO(fabian.reister): use timestamp

        // Get data.
        TIMING_START(tVisuGetData);

        TIMING_START(tRobotDescriptions);
        const description::RobotDescriptionMap robotDescriptions =
            descriptionSegment.getRobotDescriptionsLocking(timestamp);
        TIMING_END_STREAM(tRobotDescriptions, ARMARX_DEBUG);

        TIMING_START(tGlobalPoses);
        const auto globalPoses = localizationSegment.getRobotGlobalPosesLocking(timestamp);
        TIMING_END_STREAM(tGlobalPoses, ARMARX_DEBUG);

        TIMING_START(tRobotFramePoses);
        const auto frames = localizationSegment.getRobotFramePosesLocking(timestamp);
        TIMING_END_STREAM(tRobotFramePoses, ARMARX_DEBUG);

        TIMING_START(tSensorValues);
        const auto sensorValues =
            proprioceptionSegment.getSensorValuesLocking(
                timestamp, debugObserver ? &*debugObserver : nullptr);
        TIMING_END_STREAM(tSensorValues, ARMARX_DEBUG);

        TIMING_END_STREAM(tVisuGetData, ARMARX_DEBUG);


        // Build layers.
        TIMING_START(tVisuBuildLayers);

        // We need all 3 information:
        // - robot description
        // - global pose
        // - joint positions
        // => this is nothing but an armem::Robot
        ARMARX_DEBUG << "Combining robot ..."
                     << "\n- " << robotDescriptions.size() << " descriptions"
                     << "\n- " << globalPoses.size() << " global poses"
                     << "\n- " << sensorValues.size() << " joint positions";

        const robot::Robots robots =
            combine(robotDescriptions, globalPoses, sensorValues, timestamp);

        ARMARX_DEBUG << "Visualize " << robots.size() << " robots ...";
        viz::Layer layer = arviz.layer("Robots");
        visualizeRobots(layer, robots);
        std::vector layers{layer};

        ARMARX_DEBUG << "Visualize frames ...";
        if(p.framesEnabled)
        {
            viz::Layer layerFrames = arviz.layer("Frames");
            visualizeFrames(layerFrames, frames);
            layers.push_back(layerFrames);
        }

        TIMING_END_STREAM(tVisuBuildLayers, ARMARX_DEBUG);

        {
            viz::Layer layerFrames = arviz.layer("FramesIndividual");
            visualizeFramesIndividual(layerFrames, frames);
            layers.push_back(layerFrames);
        }

        if (p.forceTorque.enabled)
        {
            viz::Layer layerFrames = arviz.layer("ForceTorque");
            for (const robot::Robot& robot : robots)
            {
                const std::string& name = robot.description.name;
                if (robotNameHelper.find(name) == robotNameHelper.end())
                {
                    const std::filesystem::path robotPath = robot.description.xml.toSystemPath();
                    robotNameHelper[name] = RobotNameHelper::Create(robotPath);
                    robotModels[name]
                        = VirtualRobot::RobotIO::loadRobot(robotPath, VirtualRobot::RobotIO::eStructure);
                }

                auto model = robotModels.at(name);
                model->setJointValues(robot.config.jointMap);
                model->setGlobalPose(robot.config.globalPose.matrix());

                const proprioception::ForceTorqueValuesMap& forceTorques
                    = sensorValues.at(name).forceTorqueValuesMap;

                for (const auto& [side, ft] : forceTorques)
                {
                    ARMARX_CHECK(side == RobotNameHelper::LocationLeft or side == RobotNameHelper::LocationRight) << side;

                    const std::string forceTorqueSensorName =
                        robotNameHelper.at(name)->getArm(side).getForceTorqueSensor();

                    const Eigen::Matrix4f forceTorqueSensorPose
                        = model->getSensor(forceTorqueSensorName)->getGlobalPose();

                    const std::string xyz = "XYZ";

                    const Eigen::Vector3f from = simox::math::position(forceTorqueSensorPose);
                    for (int i = 0; i < 3; ++i)
                    {
                        simox::Color color = simox::Color((255 * Eigen::Matrix3i::Identity().col(i)).eval());
                        color.a = 128;

                        // Force arrows.
                        const float length = p.forceTorque.forceScale * ft.force(i);                        
                        const float width = std::min(
                            simox::math::rescale(std::abs(length), 0.F, 100.F, 1.F, 10.F),
                            10.F);

                        const Eigen::Vector3f to = from + length * simox::math::orientation(forceTorqueSensorPose).col(i);

                        std::stringstream key;
                        key << side << " Force " << xyz.at(i);
                        layerFrames.add(viz::Arrow(key.str()).fromTo(from, to).color(color).width(width));

                        // Torque circle arrows.
                    }
                }

            }
            layers.push_back(layerFrames);
        }


        // Commit layers.

        ARMARX_DEBUG << "Commit visualization ...";
        TIMING_START(tVisuCommit);
        arviz.commit(layers);
        TIMING_END_STREAM(tVisuCommit, ARMARX_DEBUG);

        TIMING_END_STREAM(tVisuTotal, ARMARX_DEBUG);

        if (debugObserver.has_value())
        {
            const std::string p = "Visu | ";
            debugObserver->setDebugObserverDatafield(p + "t Total (ms)", tVisuTotal.toMilliSecondsDouble());
            debugObserver->setDebugObserverDatafield(p + "t 1 Get Data (ms)", tVisuGetData.toMilliSecondsDouble());
            debugObserver->setDebugObserverDatafield(p + "t 1.1 Descriptions (ms)", tRobotDescriptions.toMilliSecondsDouble());
            debugObserver->setDebugObserverDatafield(p + "t 1.2 Global Poses (ms)", tGlobalPoses.toMilliSecondsDouble());
            debugObserver->setDebugObserverDatafield(p + "t 1.3 Frames (ms)", tRobotFramePoses.toMilliSecondsDouble());
            debugObserver->setDebugObserverDatafield(p + "t 1.4 Sensor Values (ms)", tSensorValues.toMilliSecondsDouble());
            debugObserver->setDebugObserverDatafield(p + "t 2 Build Layers (ms)", tVisuBuildLayers.toMilliSecondsDouble());
            debugObserver->setDebugObserverDatafield(p + "t 3 Commit (ms)", tVisuCommit.toMilliSecondsDouble());
        }
    }

} // namespace armarx::armem::server::robot_state
