/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI
 * @author     Christoph Pohl ( christoph dot pohl at kit dot edu )
 * @date       04.11.22
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "VariantHelperFactory.h"

#include <ArmarXCore/core/application/properties/Property.h>
#include <ArmarXCore/core/application/properties/PropertyDefinitionContainer.h>

#include <RobotAPI/libraries/aron/core/data/variant/All.h>

namespace armarx::aron::component_config::products
{
    template <>
    std::string
    VariantHelper<type::Descriptor::INT>::to_string(const data::VariantPtr& ptr) const
    {
        ARMARX_TRACE;
        return simox::alg::to_string(data::Int::DynamicCast(ptr)->getValue());
    }

    template <>
    std::string
    VariantHelper<type::Descriptor::BOOL>::to_string(const data::VariantPtr& ptr) const
    {
        ARMARX_TRACE;
        return simox::alg::to_string(data::Bool::DynamicCast(ptr)->getValue());
    }

    template <>
    std::string
    VariantHelper<type::Descriptor::STRING>::to_string(const data::VariantPtr& ptr) const
    {
        ARMARX_TRACE;
        return data::String::DynamicCast(ptr)->getValue();
    }

    template <>
    std::string
    VariantHelper<type::Descriptor::FLOAT>::to_string(const data::VariantPtr& ptr) const
    {
        ARMARX_TRACE;
        return simox::alg::to_string(data::Float::DynamicCast(ptr)->getValue());
    }

    template <>
    std::string
    VariantHelper<type::Descriptor::INT_ENUM>::to_string(const data::VariantPtr& ptr) const
    {
        ARMARX_TRACE;
        // TODO: this is should use the name of the enum value
        return simox::alg::to_string(data::Int::DynamicCast(ptr)->getValue());
    }

    template <>
    std::string
    VariantHelper<type::Descriptor::DOUBLE>::to_string(const data::VariantPtr& ptr) const
    {
        ARMARX_TRACE;
        return simox::alg::to_string(data::Double::DynamicCast(ptr)->getValue());
    }

    template <>
    aron::data::VariantPtr
    products::VariantHelper<type::Descriptor::INT>::from_string(
        const std::string& string,
        const armarx::aron::Path& path) const
    {
        ARMARX_TRACE;
        int i = string.empty() ? 0 : std::stoi(string);
        return make_int(i, path);
    }

    template <>
    aron::data::VariantPtr
    products::VariantHelper<type::Descriptor::FLOAT>::from_string(
        const std::string& string,
        const armarx::aron::Path& path) const
    {
        ARMARX_TRACE;
        float f = string.empty() ? 0 : std::stof(string);
        return make_float(f, path);
    }

    template <>
    aron::data::VariantPtr
    products::VariantHelper<type::Descriptor::DOUBLE>::from_string(
        const std::string& string,
        const armarx::aron::Path& path) const
    {
        ARMARX_TRACE;
        double d = string.empty() ? 0 : std::stod(string);
        return make_double(d, path);
    }

    template <>
    aron::data::VariantPtr
    products::VariantHelper<type::Descriptor::BOOL>::from_string(
        const std::string& string,
        const armarx::aron::Path& path) const
    {
        ARMARX_TRACE;
        if (string == "true")
        {
            return make_bool(true, path);
        }
        else if (string == "false" or string.empty())
        {
            return make_bool(false, path);
        }
        throw armarx::InvalidArgumentException("Boolean string has to be either true or false");
    }

    template <>
    aron::data::VariantPtr
    products::VariantHelper<type::Descriptor::INT_ENUM>::from_string(
        const std::string& string,
        const armarx::aron::Path& path) const
    {
        // TODO: this might not work
        ARMARX_TRACE;
        return make_int(string.empty() ? 0 : std::stoi(string), path);
    }

    template <>
    aron::data::VariantPtr
    products::VariantHelper<type::Descriptor::STRING>::from_string(
        const std::string& string,
        const armarx::aron::Path& path) const
    {
        ARMARX_TRACE;
        std::string formatted_string = "";
        if (not string.empty())
        {
            formatted_string =
                armarx::FileSystemPathBuilder_ApplyFormattingAndResolveEnvAndCMakeVars(string);
        }
        return make_string(formatted_string, path);
    }

    template <>
    void
    products::VariantHelper<type::Descriptor::INT>::set_value_from_string(
        const armarx::aron::data::VariantPtr& variant,
        const std::string& string) const
    {
        ARMARX_TRACE;
        data::Int::DynamicCast(variant)->getValue() = std::stoi(string);
    }

    template <>
    void
    products::VariantHelper<type::Descriptor::FLOAT>::set_value_from_string(
        const armarx::aron::data::VariantPtr& variant,
        const std::string& string) const
    {
        ARMARX_TRACE;
        data::Float::DynamicCast(variant)->getValue() = std::stof(string);
    }

    template <>
    void
    products::VariantHelper<type::Descriptor::DOUBLE>::set_value_from_string(
        const armarx::aron::data::VariantPtr& variant,
        const std::string& string) const
    {
        ARMARX_TRACE;
        data::Double::DynamicCast(variant)->getValue() = std::stod(string);
    }

    template <>
    void
    products::VariantHelper<type::Descriptor::BOOL>::set_value_from_string(
        const armarx::aron::data::VariantPtr& variant,
        const std::string& string) const
    {
        ARMARX_TRACE;
        data::Bool::DynamicCast(variant)->getValue() = string == "true";
    }

    template <>
    void
    products::VariantHelper<type::Descriptor::STRING>::set_value_from_string(
        const armarx::aron::data::VariantPtr& variant,
        const std::string& string) const
    {
        ARMARX_TRACE;
        data::String::DynamicCast(variant)->getValue() = string;
    }

    template <>
    void
    products::VariantHelper<type::Descriptor::INT_ENUM>::set_value_from_string(
        const armarx::aron::data::VariantPtr& variant,
        const std::string& string) const
    {
        ARMARX_TRACE;
        data::Int::DynamicCast(variant)->getValue() = std::stoi(string);
    }

    template struct products::VariantHelper<type::Descriptor::INT>;
    template struct products::VariantHelper<type::Descriptor::STRING>;
    template struct products::VariantHelper<type::Descriptor::BOOL>;
    template struct products::VariantHelper<type::Descriptor::FLOAT>;
    template struct products::VariantHelper<type::Descriptor::DOUBLE>;
    template struct products::VariantHelper<type::Descriptor::INT_ENUM>;
} // namespace armarx::aron::component_config::products